package net.knarcraft.blacksmith.config;

import java.util.Arrays;

public enum GlobalSetting {

    /**
     * The base price for repairing, regardless of durability
     *
     * <p>This allows specifying a price for each item, by setting base-price.item_name.</p>
     */
    BASE_PRICE("global.base-price.default", 10.0),

    /**
     * The base price for each durability point
     *
     * <p>If natural cost, this is the cost each missing durability point will add to the cost. If not natural cost,
     * this is the cost each present durability point will add to the cost. This allows specifying a price per
     * durability point value for each item, by setting price-per-durability-point.item_name</p>
     */
    PRICE_PER_DURABILITY_POINT("global.price-per-durability-point.default", 0.005),

    /**
     * The price increase for each level of each present enchantment
     *
     * <p>This can be specified for each possible enchantment by setting enchantment-modifiers.enchantment_name</p>
     */
    ENCHANTMENT_MODIFIER("global.enchantment-cost.default", 5),

    /**
     * Whether the cost should increase for damage taken, as opposed to increase for durability present
     */
    NATURAL_COST("global.natural-cost", true);

    private final String path;
    private final String parent;
    private final Object value;

    /**
     * Instantiates a new setting
     *
     * @param path  <p>The full config path for this setting</p>
     * @param value <p>The default value of this setting</p>
     */
    GlobalSetting(String path, Object value) {
        this.path = path;
        this.value = value;
        String[] pathParts = path.split("\\.");
        this.parent = String.join(".", Arrays.copyOfRange(pathParts, 0, pathParts.length - 1));
    }

    /**
     * Gets the full config path for this setting
     *
     * @return <p>The full config path for this setting</p>
     */
    public String getPath() {
        return path;
    }

    /**
     * Gets the parent item of the defined path
     *
     * @return <p>The parent node</p>
     */
    public String getParent() {
        return parent;
    }

    /**
     * Gets the value of this setting
     *
     * @return <p>The value of this setting</p>
     */
    Object getDefaultValue() {
        return value;
    }

}
